# Copyright (c) 2021 Satlab A/S <satlab@satlab.com>
# SPDX-License-Identifier: GPL-3.0

import xmlrpc.server
import threading

from functools import partial

from gnuradio import gr, blocks, uhd, zeromq


class SRSRadio(gr.top_block):
    def __init__(self, rx_freq, rx_bw, rx_size, rx_gain,
                 tx_freq, tx_bw, tx_size, tx_gain,
                 tun, udp_bind, udp_port, clock_source, gpio_control,
                 xmlrpc_enable, xmlrpc_bind, xmlrpc_port,
                 zmq_enable, zmq_port, tsb_tag_name=''):
        gr.top_block.__init__(self, 'radio')

        # Variables
        uhd_device = 'num_send_frames=32,num_recv_frames=256'

        # Blocks
        self.usrp_source = uhd.usrp_source(
            ','.join((uhd_device, '')),
            uhd.stream_args(
                cpu_format='fc32',
                args='',
                channels=[0],
            ),
        )
        self.usrp_source.set_clock_source(clock_source, 0)
        self.usrp_source.set_center_freq(rx_freq, 0)
        self.usrp_source.set_rx_agc(False, 0)
        self.usrp_source.set_gain(rx_gain, 0)
        self.usrp_source.set_antenna('RX2', 0)
        self.usrp_source.set_bandwidth(rx_bw, 0)
        self.usrp_source.set_samp_rate(self.samp_rate)

        self.usrp_sink = uhd.usrp_sink(
            ','.join((uhd_device, '')),
            uhd.stream_args(
                cpu_format='fc32',
                args='',
                channels=[0],
            ),
            tsb_tag_name
        )
        self.usrp_sink.set_clock_source(clock_source, 0)
        self.usrp_sink.set_center_freq(tx_freq, 0)
        self.usrp_sink.set_gain(tx_gain, 0)
        self.usrp_sink.set_antenna('TX/RX', 0)
        self.usrp_sink.set_bandwidth(tx_bw, 0)
        self.usrp_sink.set_samp_rate(self.samp_rate)

        # Configure RX/TX GPIOs. TX on FPGA_TXD0 (GPIO 8), RX on FPGA_RXD0 (GPIO 9)
        if gpio_control:
            gpio_tx = 1 << 8
            gpio_rx = 1 << 9
            gpio_mask = gpio_tx | gpio_rx
            self.usrp_sink.set_gpio_attr(  # Both ATR
                'FP0', 'CTRL', gpio_tx | gpio_rx, gpio_mask)
            self.usrp_sink.set_gpio_attr(  # Both output
                'FP0', 'DDR', gpio_tx | gpio_rx, gpio_mask)
            self.usrp_sink.set_gpio_attr(  # Both off normal
                'FP0', 'ATR_0X', 0, gpio_mask)
            self.usrp_sink.set_gpio_attr(  # TX on in TX
                'FP0', 'ATR_TX', gpio_tx, gpio_mask)
            self.usrp_sink.set_gpio_attr(  # RX on in RX
                'FP0', 'ATR_RX', gpio_rx, gpio_mask)
            self.usrp_sink.set_gpio_attr(  # Both on in RXTX
                'FP0', 'ATR_XX', gpio_tx | gpio_rx, gpio_mask)

        self.socket = blocks.socket_pdu('UDP_SERVER', udp_bind, str(udp_port), 1500, False)

        # ZMQ socket for raw samples
        if zmq_enable:
            self.zmq_samples = zeromq.pub_sink(gr.sizeof_gr_complex, 1, f'tcp://*:{zmq_port}', 100, False, -1)
            self.connect(self.usrp_source, self.zmq_samples)

        # Connect transmit chain
        self.msg_connect((self.socket, 'pdus'), (self.transmitter, 'csp_in'))
        self.connect(self.transmitter, self.usrp_sink)

        # Connect receive chain
        self.msg_connect((self.receiver, 'csp_out'), (self.socket, 'pdus'))
        self.connect(self.usrp_source, self.receiver)

        # Connect TUN interface
        if tun is not None:
            self.tuntap = blocks.tuntap_pdu(tun, tx_size, True)
            self.msg_connect((self.tuntap, 'pdus'), (self.transmitter, 'ip_in'))
            self.msg_connect((self.receiver, 'ip_out'), (self.tuntap, 'pdus'))

        # Direct function mappings
        self.get_rx_frequency = self.usrp_source.get_center_freq
        self.get_rx_gain = partial(self.usrp_source.get_gain, 0)
        self.get_tx_frequency = self.usrp_sink.get_center_freq
        self.get_tx_gain = partial(self.usrp_sink.get_gain, 0)

        # Get setters and getters in self.receiver and self.transmitter and map
        # them to self.[get|set]_[rx|tx]_[attribute]
        for rxtx in [(self.receiver, 'rx'), (self.transmitter, 'tx')]:
            for getset in ["get", "set"]:
                for attr in [x for x in dir(rxtx[0]) if callable(getattr(rxtx[0], x)) and x.startswith(f"{getset}_")]:
                    setattr(self, f"{getset}_{rxtx[1]}_{attr[4:]}", getattr(rxtx[0], attr))

        # Start XMLRPC server
        if xmlrpc_enable:
            self.xmlrpc_server = xmlrpc.server.SimpleXMLRPCServer((xmlrpc_bind, xmlrpc_port),
                                                                  allow_none=True, logRequests=False)
            # Register setters and getters
            for func in [x for x in dir(self) if callable(getattr(self, x)) and (x.startswith("get_") or x.startswith("set_"))]:
                if func not in ("get_receiver", "get_transmitter"):  # not intended for remote
                    self.xmlrpc_server.register_function(getattr(self, func), func)
            self.xmlrpc_server.register_introspection_functions()
            self.xmlrpc_server_thread = threading.Thread(target=self.xmlrpc_server.serve_forever)
            self.xmlrpc_server_thread.daemon = True
            self.xmlrpc_server_thread.start()
            print(f"Serving XML-RPC on {xmlrpc_bind}:{xmlrpc_port}")

    def get_receiver(self):
        return self.receiver

    def get_transmitter(self):
        return self.transmitter

    def set_tx_frequency(self, freq):
        self.usrp_sink.set_center_freq(freq, 0)

    def set_rx_frequency(self, freq):
        self.usrp_source.set_center_freq(freq, 0)

    def set_tx_gain(self, gain):
        self.usrp_sink.set_gain(gain, 0)

    def set_rx_gain(self, gain):
        self.usrp_source.set_gain(gain, 0)
